/*
 *  Empire - A multi-player, client/server Internet based war game.
 *  Copyright (C) 1986-1998, Dave Pare, Jeff Bailey, Thomas Ruschak,
 *                           Ken Stevens, Steve McClure
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ---
 *
 *  See the "LEGAL", "LICENSE", "CREDITS" and "README" files for all the
 *  related information and legal notices. It is expected that any future
 *  projects/authors will amend these files as needed.
 *
 *  ---
 *
 *  main.c: Thread and signal initialization for Empire Server
 * 
 *  Known contributors to this file:
 *     Dave Pare, 1994
 *     Steve McClure, 1996, 1998
 *     Doug Hay, 1998
 */

#if defined(aix) || defined(linux)
#include <unistd.h>
#endif /* aix or linux */

#include <signal.h>
#include <errno.h>
#if !defined(_WIN32)
#include <sys/ioctl.h>
#endif
#include <fcntl.h>
#include <stdio.h>

#if defined(_WIN32)
#include <winsock.h>
#endif

#include "misc.h"
#include "nat.h"
#include "file.h"
#include "player.h"
#include "empthread.h"
#include "plane.h"
#include "nuke.h"
#include "land.h"
#include "ship.h"
#include "sect.h"
#include "var.h"
#include "product.h"
#include "optlist.h"
#include "global.h"
#include "prototypes.h"

s_char	program[] = "server";

extern	void player_accept();
extern	void player_kill_idle();
extern	void update_sched();
extern  void delete_lostitems();
void nullify_objects(void);
void init_files(void);
void close_files(void);

#if defined(_WIN32)
static void loc_NTInit(void);
static void loc_NTTerm(void);
#endif

extern void mobility_check();

extern void market_update();

int
main(argc, argv)
int	argc;
char	*argv[];
{
    time_t	now;
    int  	hour[2];
    int	flags = 0;
    int	debug = 0;
    int	op;
    char *config_file = NULL;
    extern s_char *datadir;
    extern char *optarg;
    s_char tbuf[256];
#ifdef POSIXSIGNALS
    struct sigaction act;
#endif /* POSIXSIGNALS */
    
#if !defined(_WIN32)
    while ((op = getopt(argc, argv, "D:de:psh")) != EOF) {
	switch (op) {
	case 'D':
	    datadir = optarg;
	    break;
	case 'd':
	    debug++;
	    break;
	case 'e':
	    config_file = optarg;
	    break;
	case 'p':
	    flags |= EMPTH_PRINT;
	    break;
	case 's':
	    flags |= EMPTH_PRINT|EMPTH_STACKCHECK;
	    break;
	case 'h':
	default:
	    printf("Usage: %s -d -p -s\n",argv[0]);
	    return 0;
	}
    }
#endif
    if (config_file == NULL) {
	sprintf (tbuf, "%s/econfig", datadir);
	config_file = tbuf;
    }
    
    logerror("------------------------------------------------------");
#if !defined(_WIN32)
    logerror("Empire server (pid %d) started", getpid());
#else
    logerror("Empire server started");
#endif /* _WIN32 */

#if defined(_WIN32)
    loc_NTInit();
#endif
    emp_config (config_file);
    
    nullify_objects();
    
#if !defined(_WIN32)
    /* signal() should not be used with mit pthreads. Anyway if u
       have a posix threads u definitly have posix signals -- Sasha */
#if defined (POSIXSIGNALS) || defined (_EMPTH_POSIX)
#ifdef SA_SIGINFO
    act.sa_flags = SA_SIGINFO;
#endif
    sigemptyset (&act.sa_mask);
    if (debug == 0 && flags == 0) {
	disassoc();
    }
    act.sa_handler = shutdwn;
    sigaction (SIGUSR1, &act, NULL);
    sigaction (SIGTERM, &act, NULL);
    sigaction (SIGINT, &act, NULL);
    act.sa_handler = panic;
    sigaction (SIGBUS, &act, NULL);
    sigaction (SIGSEGV, &act, NULL);
    sigaction (SIGILL, &act, NULL);
    sigaction (SIGFPE, &act, NULL);
    act.sa_handler = SIG_IGN;
    sigaction(SIGPIPE, &act, NULL);
#else
    if (debug == 0 && flags == 0) {
	disassoc();
	signal(SIGUSR1, shutdwn);
	signal(SIGTERM, shutdwn);
	signal(SIGBUS, panic);
	signal(SIGSEGV, panic);
	signal(SIGILL, panic);
	signal(SIGFPE, panic);
	signal(SIGINT, shutdwn);
    }
    signal(SIGPIPE, SIG_IGN);
#endif /* POSIXSIGNALS */
#endif /* _WIN32 */
    empth_init((char **)&player, flags);
    time(&now);
#if !defined(_WIN32)
    srandom(now);
#else
    srand(now);
#endif /* _WIN32 */
    setfdtablesize(32, 256);
    global_init();
    shutdown_init();
    player_init();
    ef_init();
    init_files();
    io_init();
    gamehours(now, hour);
    
    empth_create(PP_ACCEPT, player_accept, 8192, flags, "accept",
		 "Accept network connections", 0);
    empth_create(PP_KILLIDLE, player_kill_idle, 8192, flags, "killIdle",
		 "Kills idle player connections", 0);
    empth_create(PP_SCHED, update_sched, 8192, flags, "updateSched",
		 "Schedules updates to occur", 0);
    empth_create(PP_TIMESTAMP, delete_lostitems, 8192, flags, "DelItems",
		 "Deletes old lost items", 0);
    if (opt_MOB_ACCESS) {
      /* Start the mobility access check thread */
      empth_create(PP_TIMESTAMP, mobility_check, 8192, flags, "MobCheck",
		   "Writes the timestamp file", 0);
    }

    if (opt_MARKET) {
	empth_create(PP_TIMESTAMP, market_update, 8192, flags, "MarkUpd",
		     "Updates the market", 0);
    }
#ifdef _EMPTH_LWP
    empth_yield();
    close_files(); 
#else
    empth_exit(); 
#endif	
    
#if defined(_WIN32)
    loc_NTTerm();
#endif    
    return 0;
}

void
init_files(void)
{
    ef_open(EF_NATION, O_RDWR, EFF_MEM);
    ef_open(EF_SECTOR, O_RDWR, EFF_MEM);
    ef_open(EF_SHIP, O_RDWR, EFF_MEM);
    ef_open(EF_PLANE, O_RDWR, EFF_MEM);
    ef_open(EF_LAND, O_RDWR, EFF_MEM);
    ef_open(EF_NEWS, O_RDWR, 0);
    ef_open(EF_LOAN, O_RDWR, 0);
    ef_open(EF_TREATY, O_RDWR, 0);
    ef_open(EF_NUKE, O_RDWR, EFF_MEM);
    ef_open(EF_POWER, O_RDWR, 0);
    ef_open(EF_TRADE, O_RDWR, 0);
    ef_open(EF_MAP, O_RDWR, EFF_MEM);
    ef_open(EF_BMAP, O_RDWR, EFF_MEM);
    ef_open(EF_COMM, O_RDWR, 0);
    ef_open(EF_LOST, O_RDWR, 0);
}

void
close_files(void)
{
    ef_close(EF_NATION);
    ef_close(EF_SECTOR);
    ef_close(EF_SHIP);
    ef_close(EF_PLANE);
    ef_close(EF_LAND);
    ef_close(EF_NEWS);
    ef_close(EF_LOAN);
    ef_close(EF_TREATY);
    ef_close(EF_NUKE);
    ef_close(EF_POWER);
    ef_close(EF_TRADE);
    ef_close(EF_MAP);
    ef_close(EF_COMM);
    ef_close(EF_BMAP);
    ef_close(EF_LOST);
}

/* we're going down.  try to close the files at least */
emp_sig_t
panic(int sig)
{
#if !defined(_WIN32)
#ifdef POSIXSIGNALS
    struct sigaction act;
    
    act.sa_flags = 0;
    sigemptyset (&act.sa_mask);
    act.sa_handler = SIG_DFL;
    sigaction (SIGBUS, &act, NULL);
    sigaction (SIGSEGV, &act, NULL);
    sigaction (SIGILL, &act, NULL);
    sigaction (SIGFPE, &act, NULL);
#else
    signal(SIGBUS, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);
    signal(SIGILL, SIG_DFL);
    signal(SIGFPE, SIG_DFL);
#endif /* POSIXSIGNALS */
#endif /* _WIN32 */
    logerror("server received fatal signal %d", sig);
    log_last_commands();
    close_files();
    _exit(0);
}

/* got a USR1 -- a shutdown */
emp_sig_t
shutdwn(int sig)
{
    struct player *p;
    time_t	now;
    int	running = 0;
    
    for (p = player_next(0); p != 0; p = player_next(p)) {
	if (p->state != PS_PLAYING)
	    continue;
	pr_flash(p, "Server shutting down...\n");
	if (p->command) {
	    p->state = PS_SHUTDOWN;
	    pr_flash(p, "Shutdown aborting command\n");
	    p->aborted++;
	    empth_wakeup(p->proc);
	    running++;
	}
    }
    time(&now);
    if (running) {
	/* sleep a few, wait for aborts to take effect */
	empth_sleep(now + 2);
    }
    for (p = player_next(0); p != 0; p = player_next(p)) {
	p->state = PS_KILL;
	p->aborted++;
	empth_terminate(p->proc);
	p = player_delete(p);
    }
    if (sig)
	logerror("Server shutting down on signal %d", sig);
    else
	logerror("Server shutting down at Deity's request");
    close_files();
    _exit(0);
}
void
nullify_objects(void)
{
    int i, j;
    
    if (opt_BIG_CITY) {
	dchr[SCT_CAPIT].d_flg = bigcity_dchr.d_flg;
	dchr[SCT_CAPIT].d_pkg = bigcity_dchr.d_pkg;
	dchr[SCT_CAPIT].d_build = bigcity_dchr.d_build;
	dchr[SCT_CAPIT].d_lcms = bigcity_dchr.d_lcms;
	dchr[SCT_CAPIT].d_hcms = bigcity_dchr.d_hcms;	
	dchr[SCT_CAPIT].d_name = bigcity_dchr.d_name;
    }
    for (i = 0; i < pln_maxno; i++) {
	if (opt_NO_HCMS)
	    plchr[i].pl_hcm = 0;
	if (opt_NO_LCMS)
	    plchr[i].pl_lcm = 0;
	if (opt_NO_OIL)
	    plchr[i].pl_fuel = 0;
	if (opt_PLANENAMES)
  	    plchr[i].pl_name = plchr[i].pl_planename;
    }
    for (i = 0; i < lnd_maxno; i++) {
	if (opt_NO_HCMS)
	    lchr[i].l_hcm = 0;
	if (opt_NO_LCMS)
	    lchr[i].l_lcm = 0;
    }
    for (i = 0; i < shp_maxno; i++) {
	if (opt_NO_HCMS)
	    mchr[i].m_hcm = 0;
	if (opt_NO_LCMS)
	    mchr[i].m_lcm = 0;
	if (opt_NO_OIL) {
	    if (mchr[i].m_flags & M_OIL)
		mchr[i].m_name = 0;
	}
    }
    for (i = 0; i < nuk_maxno; i++) {
	if (opt_NO_HCMS)
	    nchr[i].n_hcm = 0;
	if (opt_NO_LCMS)
	    nchr[i].n_lcm = 0;
    }
    for (i = 0; i < sct_maxno; i++) {
	if (opt_NO_HCMS)
	    dchr[i].d_hcms = 0;
	if (opt_NO_LCMS)
	    dchr[i].d_lcms = 0;
    }
    for (i = 0; i < prd_maxno; i++) {
	for (j = 0; j < pchr[i].p_nv; j++) {
	    if (opt_NO_HCMS && pchr[i].p_vtype[j] == V_HCM)
		pchr[i].p_vamt[j] = 0;
	    if (opt_NO_LCMS && pchr[i].p_vtype[j] == V_LCM)
		pchr[i].p_vamt[j] = 0;
	    if (opt_NO_OIL && pchr[i].p_vtype[j] == V_OIL)
		pchr[i].p_vamt[j] = 0;
	}
    }
}

#if defined(_WIN32)
static void
loc_NTInit()
{
    int rc;
    WORD wVersionRequested;
    WSADATA wsaData;

    wVersionRequested = MAKEWORD(2, 0);
    rc = WSAStartup(wVersionRequested, &wsaData);
    if (rc != 0) {
        logerror("WSAStartup failed.  %d", rc);
        _exit(1);
    }
}
#endif

#if defined(_WIN32)
static void
loc_NTTerm()
{
    WSACleanup();
}
#endif
